from pathlib import Path
from functools import partial

import joblib
import numpy as np
from joblib import Parallel, delayed
from sklearn.pipeline import Pipeline
from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier
from sklearn.preprocessing import Normalizer
from sklearn.feature_selection import SelectKBest, VarianceThreshold

try:
    from . import utils as ul
    from . import metrics as mt
except ImportError:
    import utils as ul
    import metrics as mt
 

np.seterr(all='ignore')

def grid_search1(clf):
    grid = GridSearchCV(SVC(random_state=1), cv=5, n_jobs=-1, param_grid=param_grid)



def grid_search(x, y, param_grid):
    grid = GridSearchCV(SVC(random_state=1), cv=5, n_jobs=-1, param_grid=param_grid)
    clf = grid.fit(x, y)
    C, gamma = clf.best_params_['C'], clf.best_params_['gamma']
    return C, gamma
    
def train(x, y, clf, out):
    model = clf.fit(x, y)
    joblib.dump(model, out)

# its discarded
def batch_train(in_dir, out_dir, C, gamma, n_job):

    def process_func(file, C, gamma):
        _, (x, y) = ul.load_data(file, normal=True)
        model = train(x, y, C, gamma)
        return model
    train_func = partial(process_func, C=C, gamma=gamma)
    dirs = Path(in_dir)
    out_dir = Path(out_dir)
    with Parallel(n_jobs=n_job) as train_pa, Parallel(n_jobs=int(n_job/2),
                                                       backend='threading') as save_pa:
        for type_ in dirs.iterdir():
            models = train_pa(delayed(train_func)(file) for file in type_.iterdir())
            type_name = type_.name
            type_dir = out_dir / type_name
            type_dir.mkdir(exist_ok=True)
            save_pa(delayed(save_model)(m, type_dir/ f.stem) for m,f in zip(models, type_.iterdir()))

def predict(x, model):
    y_pred = model.predict(x)
    try:
        y_prob = model.predict_proba(x) ## 模型训练 probability=True才能用
    except:
        y_prob = None
    return y_pred, y_prob

def cv_predict(clf, x, y, cver, method="predict"):
    y_pred = cross_val_predict(clf, x, y, cv=cver, method=method)
    return y_pred

def evaluate(x, y, cv, clf):
    k = int(cv)
    if k in (-1, 1):
        metric_dic = mt.loo_metrics(clf, x, y)
    elif int(k) > 1:
        metric_dic = mt.cv_metrics(clf, x, y, cv)
    else:  ## hold out
        pass
    return metric_dic 

def batch_evaluate(in_dir, out_dir, cv, clf, n_job):
    
    def eval_(file, cv, clf):
        _, (x, y) = ul.load_data(file)
        metric_dic = evaluate(x, y, cv, clf)
        return metric_dic
        
    eval_func = partial(eval_, cv=cv, clf=clf)
    all_metric_dic = {}
    with Parallel(n_jobs=n_job) as eval_pa:
        for type_ in in_dir.iterdir():
            metrics_iter = eval_pa(delayed(eval_func)(file) for file in type_.iterdir())
            all_metric_dic[type_] = [i for i in metrics_iter]
    return all_metric_dic

def feature_select(x, y, step, cv, clf, n_jobs):
    scaler = Normalizer()
    scaler_ft = partial(scaler.fit_transform, y=y)
    selector = VarianceThreshold()
    new_x = selector.fit_transform(x)
    score_idx = selector.get_support(indices=True)
    sb = SelectKBest(k='all')
    new_data = sb.fit_transform(new_x, y)
    f_value = sb.scores_
    idx_score = [(i, v) for i, v in zip(score_idx, f_value)]
    rank_score = sorted(idx_score, key=lambda x: x[1], reverse=True)
    feature_idx = [i[0] for i in rank_score]
    evla_func =  partial(evaluate, y=y, cv=cv, clf=clf)
    feature_num = len(feature_idx)
    step_num = feature_num // step
    if step_num * step == feature_num:
        step_num = step_num + 1
    else:
        step_num = step_num + 2
    result_ls = Parallel(n_jobs=n_jobs)(
        delayed(evla_func)(
            scaler_ft(x[:, feature_idx[:i*step]])) for i in range(1, step_num))
    return result_ls, feature_idx




# -*- coding: utf-8 -*-

"""
draw.py
~~~~~~~~~~~~
"""

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
import plotly.graph_objects as go

def p_univariate_density(acc, out):
    fig, ax = plt.subplots()
    sns.set_style('whitegrid')
    ax = sns.kdeplot(acc, ax=ax)
    ax.set_xlim(40, 100)
    ax.set_xticks(range(40, 100, 10))
    ax.set_xlabel('ACC')
    plt.savefig(out, dpi=1000, bbox_inches="tight")

def p_bivariate_density(x, y, out):
    g = sns.jointplot(x, y, kind="kde", xlim=(2, 20), ylim=(0, 1))
    g.set_axis_labels("size", "Acc")
    g.ax_joint.set_xticks(range(2, 21))
    plt.savefig(out, dpi=1000, bbox_inches="tight")

def p_comparison_type(diff_size, same_size, types_label, out):
    grid = plt.GridSpec(1, 4,)
    plt.figure(figsize=(10, 10))
    ax1 = plt.subplot(grid[0, :2])  # same type
    ax2 = plt.subplot(grid[0, 2:])  # same size

    sizes_acc = diff_size[diff_size > 0]
    sizes_label = np.where(diff_size > 0)[0] + 2
    max_acc = sizes_acc.max()
    m_size_idx = np.where(sizes_acc == max_acc)[0][0] + 1
    ax1.bar(range(1, len(sizes_acc)+1), sizes_acc, color='grey')
    ax1.set_ylim(0, 1)
    ax1.set_xlabel('Size',)
    ax1.set_xticks(range(1, len(sizes_acc)+1))
    ax1.set_xticklabels(sizes_label, fontdict={'Size': 12})
    ax1.set_ylabel('Acc', fontdict={'Size': 15})
    ax1.text(m_size_idx, max_acc+0.01, f'{max_acc:.3f}', 
             ha='center', va='bottom', fontsize=10, fontweight='bold')

    types_acc = same_size[same_size > 0]
    types_label = np.array(types_label)[same_size > 0]
    m_size_idx = np.where(types_acc == max_acc)[0][0] + 1
    ax2.bar(range(1, len(types_acc)+1), types_acc, color='grey')
    ax2.set_ylim(0, 1)
    ax2.set_xlabel('Type')
    ax2.set_xticks(range(1, len(types_acc)+1))
    ax2.set_xticklabels(types_label, fontsize=8, rotation=-35)  #
    ax2.text(m_size_idx, max_acc+0.01, f'{max_acc:.3f}',
             ha='center', va='bottom', fontsize=10, fontweight='bold')
    plt.tight_layout()
    plt.savefig(out, dpi=1000, bbox_inches="tight")


# TODO
def p_acc_heat(data, vmin, vmax, xticklabels, out, **kwargs):
    mask = data <= 0
    sns.set(font_scale=0.3)
    ax = sns.heatmap(data, vmin=vmin, vmax=vmax, cmap='RdYlGn_r', 
                      mask=mask, annot=True, fmt=".1f", robust=True,
                      annot_kws={"color": "black"}, xticklabels=True)
    ax.set_ylim([0, 19])
    ax.set_xlim([0, len(xticklabels)])
    ax.set_xticklabels(xticklabels, rotation=45, ha="center",va="center")
    ax.set_yticklabels(range(2,21))
    ax.set_xlabel('Type')
    ax.set_ylabel('Cluster Size')
    plt.savefig(out, dpi=1000, bbox_inches="tight")
    plt.close()

def p_fs(x_trick, y_trick, out, **kwargs):
    ax = plt.figure().gca()
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    x_trick = x_trick[:len(y_trick)]
    plt.plot(x_trick, y_trick)
    max_acc = kwargs.get("max_acc", None)
    best_n = kwargs.get("best_n", None)
    if max_acc and best_n:
        plt.scatter(best_n, max_acc, marker='*', c='r')
        plt.text(best_n, max_acc + 0.002, f'{best_n}, {max_acc:.4f}',
                ha='center', va='bottom', fontsize=6, fontweight='bold')
    plt.savefig(out, dpi=1000)

def visual_feature(x, y, out, labels=None):
    plt.figure()
    colors = ['navy', 'turquoise', 'darkorange', 'blue', 'grey', 'red']
    labels = labels if labels else np.unique(y).astype(int)
    for color, i, target_name in zip(colors, np.unique(y), labels):
        plt.scatter(x[y == i, 0], x[y == i, 1], alpha=.8, color=color,
                    label=target_name)
    plt.legend(loc='best', shadow=False, scatterpoints=1)
    plt.savefig(out, dpi=1000)

def visual_feature_3d(x, y, out, labels=None):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    markers = ['o', '^', '1', '*', '+', 'x', '.','D']
    for i in np.unique(y).astype(int):
        x1, x2, x3 = x[y == i, 0], x[y == i, 1], x[y == i, 2]
        ax.scatter(x1, x2, x3, marker=markers[i], label=i)
    ax.legend(loc='best')
    plt.savefig(out, dpi=1000)

def sankey(label, source, target, value, out, title):
    fig = go.Figure(
        data=[
            go.Sankey(
                customdata=label,
                orientation='h',
                node = dict(
                  pad = 15,
                  thickness = 20,
                  line = dict(color = "black", width=0.5),
                  label = label,
                  hovertemplate='%{label}'
                        ),
                link = dict(
                  source = source,
                  target = target,
                  value = value
                            ),
                    ),
        ]   )
    fig.update_layout(title_text=f"Reduction alphabet of {title}", font_size=10)
    fig.write_html(f'{out}.html')



# -*- coding: utf-8 -*-
"""
    :Author: huangsh
    :Date: 19-1-14 下午3:22
    :Description:
        过早的优化是万恶之源!
        coding...
"""
import os
import json
import argparse
from pathlib import Path
from itertools import product
from collections import defaultdict

import numpy as np

try:
    from . import draw
    from . import utils as ul
    from . import compute as cp
except ImportError:
    import draw
    import utils as ul
    import compute as cp


def sub_view(args):
    type_id = ','.join(map(str, args.type))
    size = ','.join(map(str, args.size))
    values = ul.reduce_query(type_id, size) ## 获取氨基酸约化字母表
    for item in values:
        tpi, size, cluster, method = item
        info = f"type{tpi:<3}{size:<3}{cluster:<40}{method}"
        print(info)
    if args.visual:
        values.reverse()
        cluster_dic = defaultdict(list)
        tp_cluster = type_id = [i[2] for i in values]
        type_id = [i[0] for i in values]
        [cluster_dic[i[0]].append(i[2].split('-')) for i in values]
        for ti, clusters in cluster_dic.items():
            out = f'type{ti}.html'
            clusters.insert(0, list(''.join(clusters[0])))
            ul.plot_sankey(out, clusters, f'type{ti}')

def parse_view(args, sub_parser):
    parser_v = sub_parser.add_parser('view', add_help=False, prog='raatk view',
                         usage='%(prog)s -t TYPE {1,2,3,...,73,74} -s SIZE {2,3,...,18,19}')
    parser_v.add_argument('-h', '--help', action='help')
    parser_v.add_argument('-t', '--type', nargs='+', type=int, required=True, 
                          choices=list([i for i in range(1, 75)]), help='type id')
    parser_v.add_argument('-s', '--size', nargs='+', type=int, required=True, 
                          choices=list([i for i in range(2, 20)]), help='reduce size')
    parser_v.add_argument('--visual', action='store_true', help='visualization')
    parser_v.set_defaults(func=sub_view)
    view_args = parser_v.parse_args(args)
    view_args.func(view_args)

def sub_reduce(args):
    ul.exist_file(*args.file)
    cluster, type_, size = args.cluster, args.type, args.size
    if (cluster is None) and (type_ is None or size is None):
        print('Missing arguments -t/--type and -s/--size or -c/--cluster')
        return
    for file, out in zip(args.file, args.output):
        if cluster:
            ul.check_aa(cluster)
            aa = cluster.split("-")
            out = Path(out)
            ul.reduce_to_file(file, aa, out)
        elif type_ and size:
            if "-" in type_[0]:
                start, end = type_[0].split("-")
                types = map(str, range(int(start), int(end)+1))
            else:
                types = map(str, type_)
            if "-" in size[0]:
                start, end = size[0].split("-")
                sizes = map(str, range(int(start), int(end)+1))
            else:
                sizes = map(str, size)
            cluster_info = ul.reduce_query(",".join(types), ",".join(sizes))
            out = Path(out)
            out.mkdir(exist_ok=True)
            ul.batch_reduce(file, cluster_info, out)
            if args.naa:
                naa_dir = out / "naa"
                naa_dir.mkdir(exist_ok=True)
                import shutil
                shutil.copyfile(file, naa_dir / "20-ACDEFGHIKLMNPQRSTVWY.txt")
              
def parse_reduce(args, sub_parser):
    parser_r = sub_parser.add_parser('reduce', add_help=False, prog='raatk reduce')
    parser_r.add_argument('-h', '--help', action='help')
    parser_r.add_argument('file', nargs='+', help='fasta file paths')
    parser_r.add_argument('-t', '--type', nargs='+', help='type id')
    parser_r.add_argument('-s', '--size', nargs='+', help='reduce size')
    parser_r.add_argument('-c', '--cluster', help='customized cluster')
    parser_r.add_argument('-naa', action='store_true', help='natural amino acid')
    parser_r.add_argument('-o', '--output', nargs='+', help='output file or directory')
    parser_r.set_defaults(func=sub_reduce)
    reduce_args = parser_r.parse_args(args)
    reduce_args.func(reduce_args)

def sub_extract(args):
    iscount = args.count
    k, gap, lam, n_jobs = args.kmer, args.gap, args.lam, args.process
    if args.directory:
        out = Path(args.output[0])
        out.mkdir(exist_ok=True)
        ul.batch_extract(args.file, out, k, gap, lam, n_jobs=n_jobs,count=iscount)
    else:
        if args.raa:
            raa = list(args.raa) 
        else:
            raa = list(Path(args.file[0]).stem.split('-')[-1])
        xy_ls = []
        aa_ls = [''.join(aa) for aa in product(raa, repeat=k)]
        for idx, file in enumerate(args.file):
            feature_file = Path(file)
            xy = ul.extract_feature(feature_file, raa, k, gap, lam, count=iscount)
            if args.index:
                fea_idx = np.genfromtxt(args.index, delimiter='\n').astype(int)
                xy = xy[:, fea_idx]
                aa_ls = [aa_ls[i] for i in fea_idx]
            if args.label_f:
                y = np.array([[idx]]*xy.shape[0])
                xy = np.hstack([y, xy])
            xy_ls.append(xy)
        aa_ls.insert(0, 'label') if args.label_f else 0
        header = ','.join(aa_ls)
        if args.merge:
            out = args.output[0]
            seq_mtx = np.vstack(xy_ls)
            ul.write_array(Path(out), seq_mtx, header=header)
            exit()
        for idx, o in enumerate(args.output):
            ul.write_array(Path(o), xy_ls[idx], header=header)

def parse_extract(args, sub_parser):
    parser_ex = sub_parser.add_parser('extract', add_help=False, prog='raatk extract')
    parser_ex.add_argument('-h', '--help', action='help')
    parser_ex.add_argument('file', nargs='+', help='fasta files')
    parser_ex.add_argument('-d', '--directory', action='store_true', help='feature directory')
    parser_ex.add_argument('-k', '--kmer', type=int, choices=[1,2,3], 
                          required=True, help='K-tuple or k-mer value')
    parser_ex.add_argument('-g', '--gap', type=int, default=0, help='gap value')
    parser_ex.add_argument('-l', '--lam', type=int, default=0, 
                          help='lambda-correlation value')
    parser_ex.add_argument('-raa', help='reduced amino acid cluster', default="ACDEFGHIKLMNPQRSTVWY")
    parser_ex.add_argument('-idx', '--index', default=None, help='feature index')
    parser_ex.add_argument('-m', '--merge', action='store_true', help='merge feature files into one')
    parser_ex.add_argument('-o', '--output', nargs='+', required=True, help='output directory')
    parser_ex.add_argument('-p', '--process',type=int, choices=list([i for i in range(1, os.cpu_count())]),
                                 default=1, help='cpu number')
    parser_ex.add_argument('--label-f', action='store_false', help='feature label')
    parser_ex.add_argument('--count', action='store_true', help='feature count')
    parser_ex.set_defaults(func=sub_extract)
    extract_args = parser_ex.parse_args(args)
    extract_args.func(extract_args)

def sub_hpo(args):
    pass
    # a(**dict(args._get_kwargs()))
    # params = ul.param_grid(args.C, args.gamma)
    # x, y = ul.load_data(args.file, normal=True)
    # best_c, best_gamma = cp.grid_search(x, y, params)
    # print("C: %s, gamma: %s" % (best_c, best_gamma))

#TODO
def parse_hpo(args, sub_parser):
    parser = sub_parser.add_parser('hpo', add_help=False, prog='raatk hpo')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', help='feature file for hpyper-parameter optimization')
    parser.add_argument('-clf', '--clf', default='svm', choices=['svm', 'rbf', 'knn'],
                                    help='classifier selection')
    parser.add_argument('-jobs','--n_jobs', type=int, help='the number of parallel jobs to run')

    # knn = parser.add_argument_group('KNN', description='K-nearest neighbors classifier')
    # knn.add_argument('--n_neighbors', type=int, nargs='+',
    #                     help='number of neighbors [start stop step]')
    # knn.add_argument('--weights', choices=['uniform', 'distance'], default=['uniform'], 
    #                     nargs='+', help='weight function used in prediction') 
    # knn.add_argument('--algorithm', choices=['auto', 'ball_tree', 'kd_tree', 'brute'], nargs='+',
    #                     default=['auto'], help='algorithm used to compute the nearest neighbors')
    # knn.add_argument('--leaf_size', type=int, default=30,
    #                     help='leaf size passed to BallTree or KDTree')

    # svm = parser.add_argument_group('SVM', description='Parameters for SVM classifier')
    # svm.add_argument('-c', '--C-range',  nargs='+', type=int,
    #                         help='regularization parameter value range [start, stop, [num]]')
    # svm.add_argument('-g', '--gamma-range', nargs='+', type=int,
    #                         help='Kernel coefficient value range [start, stop, [num]]')
    # svm.add_argument('--kernel', nargs='+', choices=['rbf', 'linear'],
    #                         help='kernel function')
    
    parser.set_defaults(func=sub_hpo)
    args.n_jobs = 233
    print(args.n_jobs)
    # hpo_args = parser.parse_args(args)
    # hpo_args.func(hpo_args)

def clf_parser(parser):
    svm = parser.add_argument_group('SVM', description='Parameters for SVM classifier')
    svm.add_argument('-c', '--C', default=1, type=float, help='regularization parameter')
    svm.add_argument('-g', '--gamma', default='scale', help='kernel coefficient')
    svm.add_argument('-prob', '--probability', action='store_true', help='kernel coefficient')
    svm.add_argument('-k', '--kernel', choices=['rbf', 'linear'], default='rbf',
                            help='specifies the kernel type to be used in the algorithm ')  
    svm.add_argument('-dfs', '--decision_function_shape', default='ovo', choices=['ovr', 'ovo'], 
                            help='decision function shape')
    svm.add_argument('--class_weight', default='balanced', choices=['balanced'],
                                help='default: balanced')
    svm.add_argument('-rs', '--random_state', type=int, default=1, help='random state')
    knn = parser.add_argument_group('KNN', description='K-nearest neighbors classifier')
    knn.add_argument('-n', '--n_neighbors', type=int, default=5,
                        help='number of neighbors to use by default for kneighbors queries')
    knn.add_argument('-w', '--weights', choices=['uniform', 'distance'], default='uniform', 
                        help='weight function used in prediction') 
    knn.add_argument('-al', '--algorithm', choices=['auto', 'ball_tree', 'kd_tree', 'brute'], 
                        default='auto', help='algorithm used to compute the nearest neighbors')
    knn.add_argument('-lfs', '--leaf_size', type=int, default=30,
                        help='leaf size passed to BallTree or KDTree')
    knn.add_argument('-jobs','--n_jobs', type=int, default=1, 
                            help='the number of parallel jobs to run for neighbors search')
    rf = parser.add_argument_group('RF', description='Parameters for random forest classifier') 
    rf.add_argument('-trees', '--n_estimators', type=int, default=100,
                        help='the number of trees in the forest')
    rf.add_argument('-features', '--max_features', default='auto',
                        help='the number of features to consider when looking for the best split') 
    knn.add_argument('-jobs','--n_jobs', type=int, default=1, 
                            help='the number of jobs to run in parallel')
    rf.add_argument('-rs', '--random_state', type=int, default=1, help='random state')
    return parser

def sub_train(args):
    clf = ul.select_clf(args)
    _, (x, y) = ul.load_data(args.file)
    cp.train(x, y, clf, args.output)

def parse_train(args, sub_parser):
    parser = sub_parser.add_parser('train', add_help=False, prog='raatk train',
                                     conflict_handler='resolve')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', help='feature file to train')
    parser.add_argument('-clf', '--clf', default='svm', choices=['svm', 'rbf', 'knn'],
                                help='classifier selection')
    parser.add_argument('-o', '--output',required=True, help='output directory')
    parser.add_argument('-jobs','--n_jobs', type=int, default=1, 
                            help='the number of parallel jobs to run')
    clf_parser(parser)
    parser.set_defaults(func=sub_train)
    train_args = parser.parse_args(args)
    train_args.func(train_args)

def sub_predict(args):
    _, (x, _) = ul.load_data(args.file, label_exist=False)
    model = ul.load_model(args.model)
    y_pred, y_prob = cp.predict(x, model)
    if y_prob is None:
        ul.write_array(args.output, y_pred.reshape(-1,1))
    else:
        ul.write_array(args.output, y_pred.reshape(-1,1), y_prob)

def parse_predict(args, sub_parser):
    parser = sub_parser.add_parser('predict', add_help=False, prog='raatk predict')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', help='feature file to predict')
    parser.add_argument('-m', '--model', required=True, help='model to predict')
    parser.add_argument('-o', '--output', required=True, help='output directory')
    parser.set_defaults(func=sub_predict)
    predict_args = parser.parse_args(args)
    predict_args.func(predict_args)

def sub_eval(args):
    clf = ul.select_clf(args)
    if args.directory:
        out = Path(args.output)
        all_metric_dic = cp.batch_evaluate(Path(args.file), out,args.cv, clf, args.process)
        result_json = args.output + ".json"
        ul.metric_dict2json(all_metric_dic, result_json)
    else:
        _, (x, y) = ul.load_data(args.file)
        metric_dic = cp.evaluate(x, y, args.cv, clf)
        ul.save_report(metric_dic, args.output + '.txt')
        # ul.k_roc_curve_plot(metric_dic['y_true'], metric_dic['y_prob'], args.output + '.png')       

def parse_eval(args, sub_parser):
    parser = sub_parser.add_parser('eval', add_help=False, prog='raatk eval',
                                        conflict_handler='resolve')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', help='feature file to evaluate')
    parser.add_argument('-d', '--directory', action='store_true', help='feature directory to evaluate')
    parser.add_argument('-clf', '--clf', choices=['svm', 'rf', 'knn'], default='svm', 
                                help='classifier selection')
    parser.add_argument('-cv', type=int, default=5, help='cross validation fold')
    parser.add_argument('-o', '--output', required=True, help='output directory')
    parser.add_argument('-p', '--process',type=int, choices=list([i for i in range(1, os.cpu_count())]),
                                 default=1, help='cpu numbers')     
    clf_parser(parser)
    parser.set_defaults(func=sub_eval)
    eval_args = parser.parse_args(args)
    eval_args.func(eval_args)

def sub_roc(args):
    _, (x, y) = ul.load_data(args.file)
    if args.model:
        cv = None
        clf = ul.load_model(args.model)
    else:
        cv = args.cv
        clf = ul.select_clf(args)
    ul.roc_auc_save(clf, x, y, cv, args.format, args.output)

def parse_roc(args, sub_parser):
    parser = sub_parser.add_parser('roc', add_help=False, prog='raatk roc',
                                        conflict_handler='resolve')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', help='feature file for roc')
    parser.add_argument('-cv', type=int, default=5, help='cross validation fold')
    me_group = parser.add_mutually_exclusive_group()
    me_group.add_argument('-m', '--model', help='model')
    me_group.add_argument('-clf', '--clf', choices=['svm', 'rf', 'knn'], default='svm',
                                help='classifier selection')
    fmt_choices = ['eps', 'pdf', 'png', 'ps', 'raw', 'rgba', 'svg', 'txt']
    parser.add_argument('-fmt', '--format', default="png",
                            choices=fmt_choices, help='figure format')
    parser.add_argument('-o', '--output', required=True, help='output directory')
    clf_parser(parser)
    parser.set_defaults(func=sub_roc)
    roc_args = parser.parse_args(args)
    roc_args.func(roc_args)

def sub_ifs(args):
    ul.exist_file(*args.file)
    step, cv, n_jobs = args.step, args.cv, args.process
    clf = ul.select_clf(args)
    if args.mix:
        pass
    else:
        for file, out in zip(args.file, args.output): 
            names, (x, y) = ul.load_data(file, normal=False)
            result_ls, sort_idx = cp.feature_select(x, y, step, cv, clf, n_jobs)
            x_tricks = [i for i in range(0, x.shape[1], step)]
            x_tricks.append(x.shape[1])
            mean_mt = ul.mean_metric
            acc_ls = [0] + [mean_mt(i)['acc'] for i in result_ls]
            max_acc = max(acc_ls)
            best_n = x.shape[1] if max_acc==acc_ls[-1] else acc_ls.index(max_acc)*step
            draw.p_fs(x_tricks, acc_ls, out + '.png', max_acc=max_acc, best_n=best_n)
            best_x = x[:, sort_idx[:best_n]]
            best_file =  out + '_best.csv'
            header = ','.join([names[0]]+[names[i+1] for i in sort_idx[:best_n]])
            ul.write_array(best_file, y.reshape(-1, 1), best_x, header=header)
            xtricks_arr = np.array(x_tricks[:len(acc_ls)]).reshape(-1, 1)
            acc_arr = np.array(acc_ls).reshape(-1, 1)
            ul.write_array(out+".csv", xtricks_arr, acc_arr)
            ul.write_array(out+f"-{best_n}-idx.csv", sort_idx[:best_n])
            
def parse_ifs(args, sub_parser):
    parser = sub_parser.add_parser('ifs', add_help=False, prog='raatk ifs',
                                        conflict_handler='resolve')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', nargs='+', help='feature file')
    parser.add_argument('-s', '--step', default=10, type=int, help='feature file')
    parser.add_argument('-clf', '--clf', default='svm', choices=['svm', 'rf', 'knn'],
                             help='classifier selection')
    parser.add_argument('-cv', '--cv', type=int, default=5, help='cross validation fold')                             
    parser.add_argument('-o', '--output', nargs='+', required=True, help='output folder')
    parser.add_argument('-mix', action='store_true', help='feature mix')
    parser.add_argument('-p', '--process', type=int, choices=range(1, os.cpu_count()),
                                 default=1, help='cpu core number')
    clf_parser(parser)
    parser.set_defaults(func=sub_ifs)
    ifs_args = parser.parse_args(args)
    ifs_args.func(ifs_args)

def sub_plot(args):
    ul.mkdirs(args.outdir)
    with open(args.file, 'r') as f:
        re_dic = json.load(f)
        ul.eval_plot(re_dic, Path(args.outdir), fmt=args.format)

def parse_plot(args, sub_parser):
    parser = sub_parser.add_parser('plot', add_help=False, prog='raatk plot')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', help='the result json file')
    fmt_choices = ['eps', 'pdf', 'png', 'ps', 'raw', 'rgba', 'svg']
    parser.add_argument('-fmt', '--format', default="png",
                            choices=fmt_choices, help='figure format')
    # parser.add_argument('-dpi', default=1000, type=int, help='figure format')
    parser.add_argument('-o', '--outdir', required=True, help='output directory')
    parser.set_defaults(func=sub_plot)
    plot_args = parser.parse_args(args)
    plot_args.func(plot_args)

def sub_fv(args):
    n = args.n
    data = np.genfromtxt(args.file, delimiter=',', skip_header=1, dtype=float)
    rx, y = ul.feature_reduction(data[:, 1:], data[:, 0], n_components=n)
    if len(args.label) == len(np.unique(y)):
        labels = args.label
    else:
        labels = None
    out = f'{args.output}.{args.format}'
    if n == 2:
        draw.visual_feature(rx, y, out, labels=labels)
    elif n == 3:
        draw.visual_feature_3d(rx, y, out, labels=labels)

def parse_fv(args, sub_parser):
    parser = sub_parser.add_parser('fv', add_help=False, prog='raatk fv')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', help='the result json file')
    fmt_choices = ['eps', 'pdf', 'png', 'ps', 'raw', 'rgba', 'svg']
    parser.add_argument('-fmt', '--format', default="png",
                            choices=fmt_choices, help='figure format')
    # parser.add_argument('-dpi', default=1000, type=int, help='figure format')
    parser.add_argument('-lb', '--label', nargs='+', default=[], help='feature labels')
    parser.add_argument('-n', default=2, type=int, choices=[2, 3], help='n components')
    parser.add_argument('-m', '--method', default='nca', choices=[2, 3],
                                         help='reduction method')
    parser.add_argument('-o', '--output', required=True, help='output path')
    parser.set_defaults(func=sub_fv)
    fv_args = parser.parse_args(args)
    fv_args.func(fv_args)

def sub_merge(args):
    names, _ = ul.load_structured_data(args.file[0])[0]
    mix_data = ul.merge_feature_file(args.label, args.file)
    ul.write_array(args.output, mix_data, header=','.join(names))

def parse_merge(args, sub_parser):
    parser = sub_parser.add_parser('merge', add_help=False, prog='raatk merge')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', nargs='+', help='file paths')
    parser.add_argument('-l', '--label', nargs='+', type=int, help='file format')
    parser.add_argument('-o', '--output', help='output file')
    parser.set_defaults(func=sub_merge)
    merge_args = parser.parse_args(args)
    merge_args.func(merge_args)

def sub_split(args):
    ts = args.testsize
    if 0 < ts < 1:
        data_train, data_test, names = ul.split_data(args.file, ts)
        header = ','.join(names)
        ul.write_array(f"{1-ts}_{args.output}", data_train, header=header)
        ul.write_array(f"{ts}_{args.output}", data_test, header=header)
    else:
        print("error")

def parse_split(args, sub_parser):
    parser = sub_parser.add_parser('split', add_help=False, prog='raatk split')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', help='file path')
    parser.add_argument('-ts', '--testsize', type=float, help='test size')
    parser.add_argument('-o', '--output', help='output file')
    parser.set_defaults(func=sub_split)
    split_args = parser.parse_args(args)
    split_args.func(split_args)

# TODO
def sub_transfer(args):
    for file in args.file:
        ul.fmt_transfer(Path(file), args.format)

def parse_transfer(args, sub_parser):
    parser = sub_parser.add_parser('transfer', add_help=False, prog='raatk transfer')
    parser.add_argument('-h', '--help', action='help')
    parser.add_argument('file', nargs='+', help='file path')
    parser.add_argument('-fmt', '--format', choices=['arff'], default='arff',
                         help='file path')
    parser.set_defaults(func=sub_transfer)
    transfer_args = parser.parse_args(args)
    transfer_args.func(transfer_args)


class MyArgumentParser(argparse.ArgumentParser):
    def convert_arg_line_to_args(self, arg_line):
        return arg_line.split()
 

def command_parser():
    parser = MyArgumentParser(description='reduce amino acids toolkit', 
                                fromfile_prefix_chars='@', conflict_handler='resolve')
    subparsers = parser.add_subparsers(help='sub-command help')

    parser_v = subparsers.add_parser('view', add_help=False, help='view reduced amino acids alphabet')
    parser_v.set_defaults(func=parse_view)
    parser_r = subparsers.add_parser('reduce', add_help=False, help='reduce amino acid sequence')
    parser_r.set_defaults(func=parse_reduce)
    parser_ex = subparsers.add_parser('extract', add_help=False, help='extract sequence feature')
    parser_ex.set_defaults(func=parse_extract)
    parser_hpo = subparsers.add_parser('hpo', add_help=False, help='hpyper-parameter optimization')
    parser_hpo.set_defaults(func=parse_hpo)
    parser_ev = subparsers.add_parser('eval', add_help=False, help='evaluate model')
    parser_ev.set_defaults(func=parse_eval)    
    parser_t = subparsers.add_parser('train', add_help=False, help='train model')
    parser_t.set_defaults(func=parse_train)
    parser_p = subparsers.add_parser('predict', add_help=False, help='predict data')
    parser_p.set_defaults(func=parse_predict)
    parser_roc = subparsers.add_parser('roc', add_help=False, help='roc curve evaluation')
    parser_roc.set_defaults(func=parse_roc)
    parser_f = subparsers.add_parser("ifs", add_help=False, help='incremental feature selction using ANOVA')
    parser_f.set_defaults(func=parse_ifs)
    parser_p = subparsers.add_parser("plot", add_help=False, help='visualization of evaluation result')
    parser_p.set_defaults(func=parse_plot) 
    parser_p = subparsers.add_parser("fv", add_help=False, help='high-dimensional feature visualization')
    parser_p.set_defaults(func=parse_fv)
    parser_m = subparsers.add_parser('merge', add_help=False, help='merge files into one')
    parser_m.set_defaults(func=parse_merge)
    parser_s = subparsers.add_parser('split', add_help=False, help='split file data')
    parser_s.set_defaults(func=parse_split)
    parser_tra = subparsers.add_parser('transfer', add_help=False, help='transfer file format')
    parser_tra.set_defaults(func=parse_transfer)

    tmp, kown_args = parser.parse_known_args()
    try:
        tmp.func(kown_args, subparsers)
    except AttributeError:
        pass

if __name__ == '__main__':
    command_parser()



from setuptools import setup
from raatk.__init__ import __version__

setup(name='raatk',
    version=__version__,
    description='reduce amino acid toolkit',
    url='https://github.com/huang-sh/raatk/',
    author='huangsh',
    author_email='hsh-me@outlook.com',
    license='BSD 2-Clause',
    packages=['raatk'],
    install_requires=[
        'numpy>=1.16.2',
        'matplotlib>=3.1.1',
        'scikit-learn>=0.22.1',
        'seaborn>=0.9.0',
        'plotly>=4.6.0',             
        ],
    entry_points={
        'console_scripts': [
        'raatk=raatk.__main__:command_parser',
            ]
        },
    python_requires=">=3.6",
    include_package_data=True,
    zip_safe=True)



import os
import csv
import json
import sqlite3
from pathlib import Path
from itertools import chain
from functools import partial
from itertools import product
from concurrent import futures

import joblib
import numpy as np
import numpy.lib.recfunctions as rf 
import matplotlib.pyplot as plt
from sklearn.preprocessing import Normalizer, StandardScaler
from sklearn.model_selection import train_test_split
from joblib import Parallel, delayed
from sklearn.svm import SVC
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import NeighborhoodComponentsAnalysis


try:
    from . import draw
    from . import feature as fea
    from . import metrics as mt
except ImportError:
    import draw
    import feature as fea
    import metrics as mt



BASE_PATH = Path(__file__).parent
RAA_DB = BASE_PATH / 'nr_raa_data.db'
NAA = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 
       'L', 'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']


svm_params = SVC._get_param_names()
knn_params = KNeighborsClassifier._get_param_names()
rf_params = RandomForestClassifier._get_param_names()
clf_param_names = {'svm': svm_params, 'knn': knn_params, 'rf': rf_params}
clf_dic = {'svm': SVC, 'knn': KNeighborsClassifier, 'rf': RandomForestClassifier}


def check_aa(aa):
    if aa[0] == "-" or aa[-1] == "-":
        raise ValueError("amino acid cluster is wrong!")
    if "-" not in aa:
        raise ValueError("need an amino acid cluster!")
    aa = aa.strip().upper()
    cls_ls = list(aa.replace("-", "")).sort()
    if NAA.sort() != cls_ls:
        raise ValueError("amino acid cluster is wrong!")
    
# TODO - query optimization
def reduce_query(type_id, size):
    conn = sqlite3.connect(str(RAA_DB))
    cursor = conn.cursor()
    cursor.execute('select r.type_id, c.size, c.scheme, r.method from raa r \
                inner join cluster c on r.type_id=c.type_id \
                where  c.size in (%s) and r.type_id in (%s)' % (size, type_id))
    raa_clusters = cursor.fetchall()
    cursor.close()
    conn.commit()
    conn.close()
    return raa_clusters

def read_fasta(seq):
    seq_ls = []
    for line in seq:
        line = line.strip()
        if not line:
            continue
        if line[0] == '>':
            if seq_ls:
                yield descr, ''.join(seq_ls)
                seq_ls.clear()
            descr = line
        else:
            seq_ls.append(line)
    else:
        yield descr, ''.join(seq_ls)


def reduce(seqs, aa, raa=None):
    """ reduce seq based on rr
    :param seqs: seq lines, iter
    :param aa: cluster aa, list or tuple
    :param raa: representative aa, list or tuple
    :return:
    """
    if not raa:
        raa = [i.strip()[0] for i in aa]
    for i, j in zip(aa, raa):
        if j not in i:
            raise ValueError(f'raa or clustered_aa is wrong!')
    aa_dic = dict(zip(raa, aa))
    for seq in seqs:
        descr, seq = seq
        for key, val in aa_dic.items():
            if key == val:
                continue
            else:
                for ele in val:
                    seq = seq.replace(ele, key)
        yield descr, seq
    
def reduce_to_file(file, aa, output,):
    with output.open("w") as wh:
        rh = open(file, "r")
        seqs = read_fasta(rh)
        reduced_seqs = reduce(seqs, aa)
        for descr, r_seq in reduced_seqs:
            wh.write(descr)
            wh.write("\n")
            for i in range(len(r_seq) // 80 + 1):
                wh.write(r_seq[i*80:(i+1)*80])
                wh.write("\n")
            else:
                wh.write(r_seq[(i+1)*80:])
        rh.close()

def batch_reduce(file, cluster_info, out_dir):
    with futures.ThreadPoolExecutor(len(cluster_info)) as tpe:
        to_do_map = {}
        for idx, item in enumerate(cluster_info, 1):
            type_id, size, cluster, _ = item
            aa = [i[0] for i in cluster.split("-")]
            type_n = "".join(["type",str(type_id)])
            reduce_name = '-'.join([str(size), "".join(aa)]) + '.txt'
            rfile = out_dir / type_n / reduce_name
            mkdirs(rfile.parent)
            aa = cluster.split('-')
            future = tpe.submit(reduce_to_file, file, aa, rfile)
            to_do_map[future] = type_id, size, cluster
        done_iter = futures.as_completed(to_do_map)
        for i in done_iter:
            type_id, size, cluster = to_do_map[i]
            print("done %s %s %s" % (type_id, size, cluster)) 

def extract_feature(feature_file, raa, k, gap, lam, count=False):
    with open(feature_file, "r") as rh:
        seqs = read_fasta(rh)
        fea_func = partial(fea.seq_aac, raa=raa, k=k, gap=gap, lam=lam, count=count)
        seq_vec = np.array([fea_func(sq[1]) for sq in seqs])
    return seq_vec
            
# TODO - IO optimization     
def batch_extract(in_dirs, out_dir, k, gap, lam, n_jobs=1, count=False):

    def parse_filepath(in_dirs, out_dir):
        tps_iter = [Path(i).iterdir() for i in in_dirs]
        for tps in zip(*tps_iter):
            type_dir = out_dir / tps[0].name
            type_dir.mkdir(exist_ok=True)
            sizes_iter = [size.iterdir() for size in tps]
            for sfs in zip(*sizes_iter):
                szie_stem = sfs[0].stem
                out = type_dir / (szie_stem + ".csv")
                raa_str = szie_stem.split("-")[-1]
                yield out, sfs, raa_str

    def files_extract(raa, k, gap, lam, count, *files):
        xy_ls = []
        for idx, file in enumerate(files):
            xy = extract_feature(file, raa, k, gap, lam, count=count)
            y = np.array([[idx]]*xy.shape[0])
            xy = np.hstack([y, xy])
            xy_ls.append(xy)
        new_xy = np.vstack(xy_ls)
        return new_xy

    def feature2file(out, files, raa, k, gap, lam, count):
        data = files_extract(raa, k, gap, lam, count, *files)
        aa_ls = [''.join(aa) for aa in product(raa, repeat=k)]
        aa_ls.insert(0, 'label')
        header = ','.join(aa_ls)
        write_array(out, data, header=header)

    extract_fun = partial(feature2file, k=k, gap=gap, lam=lam, count=count)
    with Parallel(n_jobs=n_jobs) as pl:
        pl(delayed(extract_fun)(*paras) for paras in parse_filepath(in_dirs, out_dir))

def dic2array(result_dic, key='OA', cls=0):
    acc_ls = []  # all type acc
    type_ls = [type_id for type_id in result_dic.keys()]
    type_ls.sort(key=lambda x: int(x[4:]))
    all_score_array = np.zeros([len(type_ls), 19])
    for idx, ti in enumerate(type_ls):
        type_ = result_dic[ti]
        score_size_ls = []
        for size in range(2, 21):
            key_scores = type_.get(str(size), {key: 0}).get(key, 0)
            score = key_scores[cls] if isinstance(key_scores, list) else key_scores
            score_size_ls.append(score)
        all_score_array[idx] = score_size_ls
    return all_score_array, type_ls

def filter_type(score_metric, type_ls, filter_num=8):
    filter_scores, filter_type = [], []
    for type_score, type_id in zip(score_metric, type_ls):
        scores = [i for i in type_score if i > 0]
        if len(scores) >= 8:
            filter_scores.append(type_score)
            filter_type.append(type_id)
    return np.array(filter_scores), filter_type
   
def heatmap_txt(data, types, out):
    with open(out, "w", newline="") as f:
        fc = csv.writer(f)
        col = ["size"] + types
        fc.writerow(col)
        for row, size in zip(data, range(2, 21)):
            line = [size] + row.tolist()
            fc.writerow(line)

def eval_plot(result_dic, out, fmt, filter_num=8):
    key = 'acc'
    scores, types = dic2array(result_dic, key=key)
    f_scores, f_types = filter_type(scores, types, filter_num=filter_num)
    
    heatmap_path = out / f'{key}_heatmap.{fmt}'
    txt_path = out / f'{key}_heatmap.csv'
    heatmap_txt(scores.T, types, txt_path)
    hmd = scores.T * 100
    fhmd = f_scores.T * 100
    vmin = np.min(hmd[hmd>0]) // 10 * 10
    vmax = np.max(hmd[hmd>0]) // 10 * 10 + 10
    draw.p_acc_heat(hmd, vmin, vmax, types, heatmap_path)
    f_heatmap_path = out / f'f{filter_num}_{key}_heatmap.{fmt}'
    draw.p_acc_heat(fhmd, vmin, vmax, f_types, f_heatmap_path)
    
    f_scores_arr = f_scores[f_scores > 0]
    size_arr = np.array([np.arange(2, 21)] * f_scores.shape[0])[f_scores > 0]
    acc_size_path = out / f'acc_size_density.{fmt}'
    acc_path = out / f'acc_density.{fmt}'
    size_arr = size_arr.flatten()
    draw.p_bivariate_density(size_arr, f_scores_arr, acc_size_path)
    draw.p_univariate_density(f_scores_arr*100, acc_path)

    max_type_idx_arr, max_size_idx_arr = np.where(f_scores == f_scores.max())
    m_type_idx, m_size_idx = max_type_idx_arr[0], max_size_idx_arr[0]  # 默认第一个
    cp_path = out / f'acc_comparsion.{fmt}'
    diff_size = f_scores[m_type_idx]
    same_size = f_scores[:, m_size_idx]
    types_label = [int(i[4:]) for i in f_types]
    draw.p_comparison_type(diff_size, same_size, types_label, cp_path)
    return f_scores_arr

def parse_path(feature_folder, filter_format='csv'):
    """
    :param feature_folder: all type feature folder path
    :return:
    """
    path = os.walk(feature_folder)
    for root, dirs, file in path:
        if root == feature_folder:
            continue
        yield root, [i for i in file if i.endswith(filter_format)]

def mkdirs(directory):
    try:
        os.makedirs(directory)
    except FileExistsError:
        pass

def load_structured_data(file): ## file for data (x,y)
    if Path(str(file)).is_file():
        structured_data = np.genfromtxt(file, delimiter=',', names=True, dtype=float)
        data = rf.structured_to_unstructured(rf.repack_fields(structured_data)) 
    else:
        raise FileNotFoundError(file) # raise error
    data = data.reshape(1, -1) if len(data.shape) == 1 else data
    names = structured_data.dtype.names
    return names, data 

def normal_data(data):
    scaler = Normalizer()
    new_data = scaler.fit_transform(data)
    return new_data

def load_data(file, label_exist=True, normal=True):
    names, data = load_structured_data(file)
    if label_exist:
        x, y = data[:, 1:], data[:, 0]
    else:
        x, y = data, None
    if normal:
        x = normal_data(x)
    return names, (x, y)


def load_model(model_path):
    model = joblib.load(model_path)
    return model

def merge_feature_file(label, file):
    if label is None:
        data_ls = [load_structured_data(file)[1] for file in file]
        mix_data = np.vstack(data_ls)
    else:
        data_ls = []
        for idx, f in zip(label, file):
            data = load_structured_data(f)[1]
            data[:, 0] = idx
            data_ls.append(data)
        mix_data = np.vstack(data_ls)
    return mix_data

def write_array(file, *data, header=None):
    data = np.hstack(data)
    if header:
        np.savetxt(file, data, delimiter=",", fmt="%.6f", header=header, comments='')
    else:
        np.savetxt(file, data, delimiter=",", fmt="%.6f")
    
def split_data(file, test_size):
    names, data = load_structured_data(file)
    data_train, data_test = train_test_split(data, test_size=test_size, random_state=1)
    return data_train, data_test, names
    
def param_grid(c, g):
    c_range = np.logspace(*c, base=2)
    gamma_range = np.logspace(*g, base=2)
    params = [{'kernel': ["rbf"], 'C': c_range,
                    'gamma': gamma_range}]
    return params

def mean_metric(cv_metric_dic):
    cv = len(cv_metric_dic)
    acc, mcc = 0, 0
    precision, recall, f1_score = 0, 0, 0
    for fold, metric in cv_metric_dic.items():
        precision = np.add(precision, metric["precision"]/cv)
        recall = np.add(recall, metric["recall"]/cv)
        f1_score = np.add(f1_score, metric["f1-score"]/cv)
        acc = np.add(acc, metric["acc"]/cv)
        mcc = np.add(mcc, metric["mcc"]/cv)
    metric_dic = {
        'precision': precision.tolist(),
        'recall': recall.tolist(), 'mcc': mcc.tolist(),
        'acc': acc.tolist(),'f1-score': f1_score.tolist()}
    return metric_dic

def metric_dict2json(all_metric_dic, path):
    result_dic = {}
    naa_metric = {}
    for type_dir, cv_metric_ls in all_metric_dic.items():
        result_dic.setdefault(type_dir.name, {})
        for size_dir, cv_metric_dic in zip(type_dir.iterdir(), cv_metric_ls):
            metric_dic = mean_metric(cv_metric_dic)
            size_ = size_dir.name.split("-")[0]
            result_dic[type_dir.name][size_] = metric_dic
    naa_metric = result_dic.pop('naa', {})
    if naa_metric:
        [result_dic[tk].update(naa_metric) for tk in result_dic]
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(result_dic, f, indent=4)

def roc_auc_save(clf, x, y, cv, fmt, out):
    if cv is None:
        viz = mt.roc_curve_plot(clf, x, y)
    elif cv in (1, -1):
        viz = mt.loo_roc_curve_plot(clf, x, y)
    else:
        viz = mt.cv_roc_curve_plot(clf, x, y, cv)
    if fmt == 'txt':
        fpr = viz.fpr.reshape(-1, 1)
        tpr = viz.fpr.reshape(-1, 1)
        auc = viz.roc_auc
        write_array(f"{out}-{auc:.2f}.csv", fpr, tpr)
    else:
        plt.savefig(f"{out}.{fmt}", dpi=1000)


TEXT = """
    敏感度(Sensitivity, SN)也称召回率(Recall, RE):	
            Sn = Recall = TP / (TP + FN)
    特异性(Specificity, SP):
            Sp = TN / (TN + FP)
    精确率(Precision, PR)也称阳极预测值(Positive Predictive Value, PPV):	
            Precision= PPV = TP / (TP + FP)
    预测成功率(Accuracy, Acc):
            Acc = (TP + TN) / (TP + FP + TN + FN)
    Matthew 相关系数(Matthew's correlation coefficient, Mcc):
        MCC = (TP*TN- FP*FN)/sqrt((TP + FP)*(TN + FN)*(TP + FN)*(TN + FP)).其中sqrt代表开平方.
"""
              
def save_report(metric_dic, report_file):
    with open(report_file, "w", encoding="utf-8") as f:
        head = (f"     {'tp':^5}{'fn':^5}{'fp':^5}{'tn':^5}" +
                f"{'recall':^8}{'precision':^11}{'f1-score':^11}\n")
        for idx, sm in metric_dic.items():
            f.write(f"{idx:^50}")
            f.write('\n')
            for i, j in enumerate(sm['cm']):
                f.write(f"{i:<4}")
                f.write('  '.join(map(str, j.tolist())))
                f.write('\n')
            f.write('\n')
            f.write(head)
            tp, fn, fp, tn = sm['tp'], sm['fn'], sm['fp'], sm['tn']
            ppr, recall, f1s = sm['precision'], sm['recall'], sm['f1-score']
            cls_i = "{:^5}{:^5}{:^5}{:^5}{:^5}{:^8.2f}{:^11.2f}{:^11.2f}\n"
            acc_i = "acc{:>48.2f}\n"
            mcc_i = "mcc{:>48.2f}"
            for i in range(len(tp)):
                line = (i, tp[i], fn[i], fp[i], tn[i], recall[i], ppr[i], f1s[i])
                f.write(cls_i.format(*line))
            f.write(acc_i.format(sm['acc']))
            f.write(mcc_i.format(sm['mcc']))
            f.write("\n")
            f.write("-"*55)
            f.write("\n")
        f.write("\n")

def filter_args(kwargs, clf_params):
    params = *kwargs,
    target_params = set(params) & set(clf_params)
    param_kic = {}
    for p in target_params:
        if isinstance(kwargs[p], (int, float)):
            param_kic[p] = kwargs[p]
        elif kwargs[p].replace(".",'').isnumeric():
            param_kic[p] = float(kwargs[p])
        else:
            param_kic[p] = kwargs[p]    
    return param_kic

def select_clf(args):
    kwargs = dict(args._get_kwargs())
    param_dic = filter_args(kwargs, clf_param_names[args.clf])
    clf = clf_dic[args.clf](**param_dic)
    return clf

def exist_file(*file_path):
    for file in file_path:
        f = Path(file)
        if f.is_file():
            pass
        else:
            print("File not found!")
            exit()

#TODO add other method
def feature_reduction(x, y, n_components=2):
    from sklearn.pipeline import make_pipeline
    nca = make_pipeline(Normalizer(),
            NeighborhoodComponentsAnalysis(init='auto',
                            n_components=n_components, random_state=1))
    rx = nca.fit_transform(x,y)
    return rx, y

def fmt_transfer(file, fmt='arff'):
    names, data = load_structured_data(file)
    out = file.with_suffix(f'.{fmt}')
    with open(out, 'w', newline='\n') as f:
        fc = csv.writer(f)
        fc.writerow([r'% generated by RAATK'])
        fc.writerow(['@Relation Protein'])
        for aa in names[1:]:
            fc.writerow([f'@attribute {aa} NUMERIC'])
        labels = ','.join([f'class{int(i)}' for i in np.unique(data[:, 0])])
        classes = ''.join(['{', labels, '}'])
        f.write(f"@attribute class {classes}\n")
        fc.writerow(["@Data"])
        for row in data:
            row = row.tolist()
            line = row[1:] + [f'class{int(row[0])}']
            fc.writerow(line)

def cluster_link(source, target):
    sl, tl, vl = [], [], []
    for ti, taac in enumerate(target):
        taa_set = set(taac)
        aac_len = len(taac)
        for si, saac in enumerate(source):
            intersect = taa_set & set(saac)
            if intersect:
                sl.append(si)
                tl.append(ti)
                vl.append(len(intersect))
                aac_len -= len(intersect)
            if aac_len == 0:
                break
    return sl, tl, vl 

def type_link(clusters):
    base_idx = 0    
    source_idx, target_idx, values = [], [], []
    for i in range(len(clusters)-1):
        sl, tl, vl = cluster_link(clusters[i], clusters[i+1])
        sidx = [i+base_idx for i in sl]
        base_idx += len(clusters[i])
        tidx = [i+base_idx for i in tl]
        source_idx.extend(sidx)
        target_idx.extend(tidx)
        values.extend(vl)
    return source_idx, target_idx, values

def plot_sankey(out, clusters, title):
    source_idx, target_idx, values = type_link(clusters)
    labels = list(chain(*clusters))
    draw.sankey(labels, source_idx, target_idx, values, out, title)